from __future__ import annotations

import asyncio
import datetime
import logging
import os
import subprocess
from typing import Coroutine

import structlog
import zmq
import zmq.asyncio
from toolz.dicttoolz import dissoc

ctx = zmq.asyncio.Context()

logger = logging.getLogger("adaptive_scheduler._mock_scheduler")
logger.setLevel(logging.INFO)
log = structlog.wrap_logger(logger)

DEFAULT_URL = "tcp://127.0.0.1:60547"


class MockScheduler:
    f"""Emulates a HPC-like scheduler.

    Start an instance of `MockScheduler` and then you are able to do
    ```
    python _mock_scheduler.py --queue
    python _mock_scheduler.py --submit JOB_NAME_HERE script_here.sh  # returns JOB_ID
    python _mock_scheduler.py --cancel JOB_ID
    ```

    Parameters
    ----------
    startup_delay : int
        The waiting before starting the process.
    max_running_jobs : int
        Maximum number of simultaneously running jobs.
    refresh_interval : int
        Refresh interval of checking whether proccesses are still running.
    bash : str, default: "bash"
        ``bash`` executable.
    url : str, optional
        The URL of the socket. Defaults to {DEFAULT_URL}.
    """

    def __init__(
        self,
        startup_delay=3,
        max_running_jobs=4,
        refresh_interval=0.1,
        bash="bash",
        url=None,
    ):
        self._current_queue = {}
        self._job_id = 0
        self.max_running_jobs = max_running_jobs
        self.startup_delay = startup_delay
        self.refresh_interval = refresh_interval
        self.bash = bash
        self.ioloop = asyncio.get_event_loop()
        self.refresh_task = self.ioloop.create_task(self._refresh_coro())
        self.url = url or DEFAULT_URL
        self.command_listener_task = self.ioloop.create_task(self._command_listener())

    def queue(self, only_me: bool = True):
        # only_me doesn't do anything, but the argument is there
        # because it is in the other schedulers.

        # remove the "proc" entries because they aren't pickable
        return {
            job_id: dissoc(info, "proc") for job_id, info in self._current_queue.items()
        }

    def _queue_is_full(self):
        n_running = sum(info["state"] == "R" for info in self._current_queue.values())
        return n_running >= self.max_running_jobs

    def _get_new_job_id(self):
        job_id = self._job_id
        self._job_id += 1
        return str(job_id)

    async def _submit_coro(self, job_id: str, fname: str):
        await asyncio.sleep(self.startup_delay)
        while self._queue_is_full():
            await asyncio.sleep(self.refresh_interval)
        self._submit(job_id, fname)

    def _submit(self, job_id: str, fname: str):
        if job_id in self._current_queue:
            # job_id could be cancelled before it started
            cmd = f"{self.bash} {fname}"
            proc = subprocess.Popen(
                cmd.split(), stdout=subprocess.PIPE, env=dict(os.environ, JOB_ID=job_id)
            )
            info = self._current_queue[job_id]
            info["proc"] = proc
            info["state"] = "R"

    def submit(self, job_name: str, fname: str):
        job_id = self._get_new_job_id()
        self._current_queue[job_id] = {
            "job_name": job_name,
            "proc": None,
            "state": "P",
            "timestamp": str(datetime.datetime.now()),
        }
        self.ioloop.create_task(self._submit_coro(job_id, fname))
        return job_id

    def cancel(self, job_id: str):
        job_id = str(job_id)
        info = self._current_queue.pop(job_id)
        if info["proc"] is not None:
            info["proc"].kill()

    async def _refresh_coro(self):
        while True:
            try:
                await asyncio.sleep(self.refresh_interval)
                self._refresh()
            except Exception as e:
                print(e)

    def _refresh(self):
        for job_id, info in self._current_queue.items():
            if info["state"] == "R" and info["proc"].poll() is not None:
                info["state"] = "F"

    async def _command_listener(self) -> Coroutine:
        log.debug("started _command_listener")
        socket = ctx.socket(zmq.REP)
        socket.bind(self.url)
        try:
            while True:
                request = await socket.recv_pyobj()
                reply = self._dispatch(request)
                await socket.send_pyobj(reply)
        finally:
            socket.close()

    def _dispatch(self, request: tuple[str, ...]):
        log.debug("got a request", request=request)
        request_type, *request_arg = request
        try:
            if request_type == "submit":
                job_name, fname = request_arg
                log.debug("submitting a task", fname=fname, job_name=job_name)
                job_id = self.submit(job_name, fname)
                return job_id
            elif request_type == "cancel":
                job_id = request_arg[0]
                log.debug("got a cancel request", job_id=job_id)
                self.cancel(job_id)
                return None
            elif request_type == "queue":
                log.debug("got a queue request")
                return self._current_queue
            else:
                log.debug("got unknown request")
        except Exception as e:
            return e


def _external_command(command: tuple[str, ...], url: str):
    async def _coro(command, url) -> None:
        with ctx.socket(zmq.REQ) as socket:
            socket.setsockopt(zmq.RCVTIMEO, 2000)
            socket.connect(url)
            await socket.send_pyobj(command)
            reply = await socket.recv_pyobj()
            return reply

    coro = _coro(command, url)
    ioloop = asyncio.get_event_loop()
    task = ioloop.create_task(coro)
    return ioloop.run_until_complete(task)


def queue(url: str = DEFAULT_URL):
    return _external_command(("queue",), url)


def submit(job_name: str, fname: str, url: str = DEFAULT_URL) -> None:
    return _external_command(("submit", job_name, fname), url)


def cancel(job_id: str, url: str = DEFAULT_URL) -> None:
    return _external_command(("cancel", job_id), url)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--queue", action="store_true")
    parser.add_argument("--submit", action="store", nargs=2, type=str)
    parser.add_argument("--cancel", action="store", type=str)
    parser.add_argument("--url", action="store", type=str, default=DEFAULT_URL)
    args = parser.parse_args()

    if args.queue:
        print(queue(args.url))
    elif args.submit:
        job_name, fname = args.submit
        print(submit(job_name, fname, args.url))
    elif args.cancel:
        job_id = args.cancel
        cancel(job_id, args.url)
        print("Cancelled")



from __future__ import annotations

import sys
import time
import warnings
from copy import copy

from adaptive.learner.base_learner import BaseLearner

from adaptive_scheduler.utils import _print_same_line

inf = sys.float_info.max


def ensure_hashable(x):
    try:
        hash(x)
        return x
    except TypeError:
        msg = "The items in `sequence` need to be hashable, {}. Make sure you reflect this in your function."
        if isinstance(x, dict):
            warnings.warn(msg.format("we converted `dict` to `tuple(dict.items())`"))
            return tuple(x.items())
        else:
            warnings.warn(msg.format("we tried to cast the items to a tuple"))
            return tuple(x)


class SequenceLearner(BaseLearner):
    def __init__(self, function, sequence):
        warnings.warn("Use `adaptive.SequenceLearner` from adaptive>=0.9!")
        print(
            "I am going to sleep for 10 seconds to annoy you into updating your code!"
        )
        for i in range(10, -1, -1):
            _print_same_line(f"Sleeping for {i} seconds more.", new_line_end=(i == 0))
            time.sleep(1)
        self.function = function
        self._to_do_seq = {ensure_hashable(x) for x in sequence}
        self._npoints = len(sequence)
        self.sequence = copy(sequence)
        self.data = {}
        self.pending_points = set()

    def ask(self, n, tell_pending=True):
        points = []
        loss_improvements = []
        for point in self._to_do_seq:
            if len(points) >= n:
                break
            points.append(point)
            loss_improvements.append(inf / self._npoints)

        if tell_pending:
            for p in points:
                self.tell_pending(p)

        return points, loss_improvements

    def _get_data(self):
        return self.data

    def _set_data(self, data):
        if data:
            self.tell_many(*zip(*data.items()))

    def loss(self, real=True):
        if not (self._to_do_seq or self.pending_points):
            return 0
        else:
            npoints = self.npoints + (0 if real else len(self.pending_points))
            return inf / npoints

    def remove_unfinished(self):
        for p in self.pending_points:
            self._to_do_seq.add(p)
        self.pending_points = set()

    def tell(self, point, value):
        self.data[point] = value
        self.pending_points.discard(point)
        self._to_do_seq.discard(point)

    def tell_pending(self, point):
        self.pending_points.add(point)
        self._to_do_seq.discard(point)

    def done(self):
        return not self._to_do_seq and not self.pending_points

    def result(self):
        """Get back the data in the same order as ``sequence``."""
        if not self.done():
            raise Exception("Learner is not yet complete.")
        return [self.data[ensure_hashable(x)] for x in self.sequence]

    @property
    def npoints(self):
        return len(self.data)



from __future__ import annotations

import asyncio
from collections import defaultdict
from contextlib import suppress
from datetime import datetime, timedelta
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from ipywidgets import (
    HTML,
    Button,
    Checkbox,
    Dropdown,
    HBox,
    Layout,
    Text,
    Textarea,
    VBox,
)

if TYPE_CHECKING:
    from adaptive_scheduler.server_support import RunManager


def _get_fnames(run_manager, only_running: bool) -> list[Path]:
    if only_running:
        fnames = []
        for entry in run_manager.database_manager.as_dicts():
            if entry["is_done"]:
                continue
            if entry["log_fname"] is not None:
                fnames.append(entry["log_fname"])
            fnames += entry["output_logs"]
        return sorted(map(Path, fnames))
    pattern = f"{run_manager.job_name}-*"
    logs = set(Path(run_manager.scheduler.log_folder).glob(pattern))
    logs |= set(Path(".").glob(pattern))
    return sorted(logs)


def _failed_job_logs(fnames, run_manager, only_running):

    running = {
        Path(e["log_fname"]).stem
        for e in run_manager.database_manager.as_dicts()
        if e["log_fname"] is not None
    }

    fnames = {
        fname.stem for fname in fnames if fname.suffix != run_manager.scheduler.ext
    }
    failed = fnames - running
    failed = [Path(f) for stem in failed for f in glob(f"{stem}*")]

    def maybe_append(fname: str, other_dir: Path, lst: list[Path]):
        p = Path(fname)
        p_other = other_dir / p.name
        if p.exists():
            lst.append(p)
        elif p_other.exists():
            lst.append(p_other)

    if not only_running and run_manager.move_old_logs_to is not None:
        base = Path(run_manager.move_old_logs_to)
        for e in run_manager.database_manager.failed:
            if not e["is_done"]:
                for f in e["output_logs"]:
                    maybe_append(f, base, failed)
            maybe_append(e["log_fname"], base, failed)
    return failed


def _files_that_contain(fnames: list[Path], text: str):
    def contains(fname: Path, text: str):
        with fname.open("r", encoding="utf-8") as f:
            for line in f:
                if text in line:
                    return True
            return False

    return [fname for fname in fnames if contains(fname, text)]


def _sort_fnames(sort_by, run_manager, fnames):
    def _try_transform(f):
        def _f(x):
            try:
                return f(x)
            except Exception:
                return x

        return _f

    def _sort_key(value):
        x, fname = value
        if isinstance(x, str):
            return -1, fname
        return float(x), fname

    mapping = {
        "Alphabetical": (None, lambda x: ""),
        "CPU %": ("cpu_usage", lambda x: f"{x:.1f}%"),
        "Mem %": ("mem_usage", lambda x: f"{x:.1f}%"),
        "Last editted": (
            "timestamp",
            lambda x: f"{(np.datetime64(datetime.now()) - x) / 1e9}s ago",
        ),
        "Loss": ("latest_loss", lambda x: f"{x:.2f}"),
        "npoints": ("npoints", lambda x: f"{x} pnts"),
        "Elapsed time": ("elapsed_time", lambda x: f"{x / 1e9}s"),
    }

    def extract(df, fname, key):
        df_sel = df[df.log_fname.str.contains(fname.name)]
        values = df_sel[key].values
        if values:
            return values[0]
        else:
            return "?"

    if sort_by != "Alphabetical":
        fname_mapping = defaultdict(list)
        for fname in fnames:
            fname_mapping[fname.stem].append(fname)

        df = run_manager.parse_log_files()
        if df.empty:
            return fnames
        log_fnames = set(df.log_fname.apply(Path))
        df_key, transform = mapping[sort_by]
        stems = [fname.stem for fname in log_fnames]
        vals = [extract(df, fname, df_key) for fname in log_fnames]
        val_stem = sorted(zip(vals, stems), key=_sort_key, reverse=True)

        result = []
        for val, stem in val_stem:
            val = _try_transform(transform)(val)
            for fname in fname_mapping[stem]:
                result.append((f"{val}: {fname.name}", fname))

        missing = fname_mapping.keys() - set(stems)
        for stem in sorted(missing):
            for fname in fname_mapping[stem]:
                result.append((f"?: {fname.name}", fname))
        return result

    return fnames


def _read_file(fname: Path, max_lines: int = 500) -> str:
    try:
        with fname.open("r", encoding="utf-8") as f:
            lines = f.readlines()

            if len(lines) > max_lines:
                lines = lines[-max_lines:]
                lines.insert(0, f"Only displaying the last {max_lines} lines!")
                lines.insert(1, "\n")
            return "".join(lines)
    except UnicodeDecodeError:
        return f"Could not decode file ({fname})!"
    except Exception as e:
        return f"Exception with trying to read {fname}:\n{e}."


def log_explorer(run_manager) -> VBox:  # noqa: C901
    def _update_fname_dropdown(
        run_manager,
        fname_dropdown,
        only_running_checkbox,
        only_failed_checkbox,
        sort_by_dropdown,
        contains_text,
    ):
        def on_click(_):
            current_value = fname_dropdown.value
            fnames = _get_fnames(run_manager, only_running_checkbox.value)
            if only_failed_checkbox.value:
                fnames = _failed_job_logs(
                    fnames, run_manager, only_running_checkbox.value
                )
            if contains_text.value.strip() != "":
                fnames = _files_that_contain(fnames, contains_text.value.strip())
            fnames = _sort_fnames(sort_by_dropdown.value, run_manager, fnames)
            fname_dropdown.options = fnames
            with suppress(Exception):
                fname_dropdown.value = current_value
            fname_dropdown.disabled = not fnames

        return on_click

    def _last_editted(fname: Path) -> float:
        try:
            return fname.stat().st_mtime
        except FileNotFoundError:
            return -1.0

    async def _tail_log(fname: Path, textarea: Textarea) -> None:
        T = -2.0  # to make sure the update always triggers
        while True:
            await asyncio.sleep(2)
            try:
                T_new = _last_editted(fname)
                if T_new > T:
                    textarea.value = _read_file(fname, run_manager.max_log_lines)
                    T = T_new
            except asyncio.CancelledError:
                return
            except Exception:
                pass

    def _tail(
        dropdown,
        tail_button,
        textarea,
        update_button,
        only_running_checkbox,
        only_failed_checkbox,
    ):
        tail_task = None
        ioloop = asyncio.get_running_loop()

        def on_click(_):
            nonlocal tail_task
            if tail_task is None:
                fname = dropdown.options[dropdown.index]
                tail_task = ioloop.create_task(_tail_log(fname, textarea))
                tail_button.description = "cancel tail log"
                tail_button.button_style = "danger"
                tail_button.icon = "window-close"
                dropdown.disabled = True
                update_button.disabled = True
                only_running_checkbox.disabled = True
                only_failed_checkbox.disabled = True
            else:
                tail_button.description = "tail log"
                tail_button.button_style = "info"
                tail_button.icon = "refresh"
                dropdown.disabled = False
                only_running_checkbox.disabled = False
                only_failed_checkbox.disabled = False
                update_button.disabled = False
                tail_task.cancel()
                tail_task = None

        return on_click

    def _on_dropdown_change(textarea):
        def on_change(change):
            if (
                change["type"] == "change"
                and change["name"] == "value"
                and change["new"] is not None
            ):
                textarea.value = _read_file(change["new"], run_manager.max_log_lines)

        return on_change

    def _click_button_on_change(button):
        def on_change(change):
            if change["type"] == "change" and change["name"] == "value":
                button.click()

        return on_change

    fnames = _get_fnames(run_manager, only_running=False)
    # no need to sort `fnames` because the default sort_by option is alphabetical
    text = _read_file(fnames[0], run_manager.max_log_lines) if fnames else ""
    textarea = Textarea(text, layout=dict(width="auto"), rows=20)
    sort_by_dropdown = Dropdown(
        description="Sort by",
        options=["Alphabetical", "CPU %", "Mem %", "Last editted", "Loss", "npoints"],
    )
    contains_text = Text(description="Has string")
    fname_dropdown = Dropdown(description="File name", options=fnames)
    fname_dropdown.observe(_on_dropdown_change(textarea))
    only_running_checkbox = Checkbox(
        description="Only files of running jobs", indent=False
    )
    only_failed_checkbox = Checkbox(
        description="Only files of failed jobs (might include false positives)",
        indent=False,
    )
    update_button = Button(
        description="update file list", button_style="info", icon="refresh"
    )
    update_button.on_click(
        _update_fname_dropdown(
            run_manager,
            fname_dropdown,
            only_running_checkbox,
            only_failed_checkbox,
            sort_by_dropdown,
            contains_text,
        )
    )
    sort_by_dropdown.observe(_click_button_on_change(update_button))
    only_running_checkbox.observe(_click_button_on_change(update_button))
    only_failed_checkbox.observe(_click_button_on_change(update_button))
    tail_button = Button(description="tail log", button_style="info", icon="refresh")
    tail_button.on_click(
        _tail(
            fname_dropdown,
            tail_button,
            textarea,
            update_button,
            only_running_checkbox,
            only_failed_checkbox,
        )
    )
    title = HTML("<h2><tt>adaptive_scheduler.widgets.log_explorer</tt></h2>")
    return VBox(
        [
            title,
            only_running_checkbox,
            only_failed_checkbox,
            update_button,
            sort_by_dropdown,
            contains_text,
            fname_dropdown,
            tail_button,
            textarea,
        ],
        layout=Layout(border="solid 2px gray"),
    )


def _info_html(run_manager) -> str:
    queue = run_manager.scheduler.queue(me_only=True)
    run_manager.database_manager.update(queue)
    jobs = [job for job in queue.values() if job["job_name"] in run_manager.job_names]
    n_running = sum(job["state"] in ("RUNNING", "R") for job in jobs)
    n_pending = sum(job["state"] in ("PENDING", "Q", "CONFIGURING") for job in jobs)
    n_done = sum(job["is_done"] for job in run_manager.database_manager.as_dicts())
    n_failed = len(run_manager.database_manager.failed)
    n_failed_color = "red" if n_failed > 0 else "black"

    status = run_manager.status()
    color = {
        "cancelled": "orange",
        "not yet started": "orange",
        "running": "blue",
        "failed": "red",
        "finished": "green",
    }[status]

    def _table_row(i, key, value):
        """Style the rows of a table. Based on the default Jupyterlab table style."""
        style = "text-align: right; padding: 0.5em 0.5em; line-height: 1.0;"
        if i % 2 == 1:
            style += " background: var(--md-grey-100);"
        return (
            f'<tr><th style="{style}">{key}</th><th style="{style}">{value}</th></tr>'
        )

    info = [
        ("status", f'<font color="{color}">{status}</font>'),
        ("# running jobs", f'<font color="blue">{n_running}</font>'),
        ("# pending jobs", f'<font color="orange">{n_pending}</font>'),
        ("# finished jobs", f'<font color="green">{n_done}</font>'),
        ("# failed jobs", f'<font color="{n_failed_color}">{n_failed}</font>'),
        ("elapsed time", timedelta(seconds=run_manager.elapsed_time())),
    ]

    with suppress(Exception):
        df = run_manager.parse_log_files()
        t_last = (pd.Timestamp.now() - df.timestamp.max()).seconds

        overhead = df.mem_usage.mean()
        red_level = max(0, min(int(255 * overhead / 100), 255))
        overhead_color = f"#{red_level:02x}{255 - red_level:02x}{0:02x}"
        overhead_html_value = f'<font color="{overhead_color}">{overhead:.2f}%</font>'

        cpu = df.cpu_usage.mean()
        red_level = max(0, min(int(255 * cpu / 100), 255))
        cpu_color = f"#{red_level:02x}{red_level:02x}{0:02x}"
        cpu_html_value = f'<font color="{cpu_color}">{cpu:.2f}%</font>'

        from_logs = [
            ("# of points", df.npoints.sum()),
            ("mean CPU usage", cpu_html_value),
            ("mean memory usage", f"{df.mem_usage.mean().round(1)} %"),
            ("mean overhead", overhead_html_value),
            ("last log-entry", f"{t_last}s ago"),
        ]
        for key in ["npoints/s", "latest_loss", "nlearners"]:
            with suppress(Exception):
                from_logs.append((f"mean {key}", f"{df[key].mean().round(1)}"))
        msg = "this is extracted from the log files, so it might not be up-to-date"
        abbr = '<abbr title="{}">{}</abbr>'  # creates a tooltip
        info.extend([(abbr.format(msg, k), v) for k, v in from_logs])

    table = "\n".join(_table_row(i, k, v) for i, (k, v) in enumerate(info))

    return f"""
        <table>
        {table}
        </table>
    """


def info(run_manager: RunManager) -> None:
    """Display information about the `RunManager`.

    Returns an interactive ipywidget that can be
    visualized in a Jupyter notebook.
    """
    from IPython.display import display
    from ipywidgets import HTML, Button, Layout, VBox

    status = HTML(value=_info_html(run_manager))

    layout = Layout(width="200px")

    cancel_button = Button(
        description="cancel jobs", layout=layout, button_style="danger", icon="stop"
    )
    cleanup_button = Button(
        description="cleanup log and batch files",
        layout=layout,
        button_style="danger",
        icon="remove",
    )
    update_info_button = Button(
        description="update info",
        layout=layout,
        icon="refresh",
    )
    update_info_button.style.button_color = "lightgreen"
    load_learners_button = Button(
        description="load learners",
        layout=layout,
        button_style="info",
        icon="download",
    )
    show_logs_button = Button(
        description="show logs", layout=layout, button_style="info", icon="book"
    )
    widgets = {
        "update info": update_info_button,
        "cancel": HBox([cancel_button], layout=layout),
        "cleanup": HBox([cleanup_button], layout=layout),
        "load learners": load_learners_button,
        "show logs": show_logs_button,
    }

    def switch_to(box, *buttons, _callable=None):
        def on_click(_):
            box.children = tuple(buttons)
            if _callable is not None:
                _callable()

        return on_click

    box = VBox([])

    log_widget = None

    def update(_):
        status.value = _info_html(run_manager)

    def load_learners(_):
        run_manager.load_learners()

    def toggle_logs(_):
        nonlocal log_widget

        if log_widget is None:
            log_widget = log_explorer(run_manager)

        b = widgets["show logs"]
        if b.description == "show logs":
            b.description = "hide logs"
            box.children = (*box.children, log_widget)
        else:
            b.description = "show logs"
            box.children = box.children[:-1]

    def cancel():
        run_manager.cancel()
        update(None)

    def cleanup(include_old_logs):
        def _callable():
            run_manager.cleanup(include_old_logs.value)
            update(None)

        return _callable

    widgets["update info"].on_click(update)
    widgets["show logs"].on_click(toggle_logs)
    widgets["load learners"].on_click(load_learners)

    # Cancel button with confirm/deny option
    confirm_cancel_button = Button(
        description="Confirm", button_style="success", icon="check"
    )
    deny_cancel_button = Button(description="Deny", button_style="danger", icon="close")

    cancel_button.on_click(
        switch_to(widgets["cancel"], confirm_cancel_button, deny_cancel_button)
    )
    deny_cancel_button.on_click(switch_to(widgets["cancel"], cancel_button))
    confirm_cancel_button.on_click(
        switch_to(widgets["cancel"], cancel_button, _callable=cancel)
    )

    # Cleanup button with confirm/deny option
    include_old_logs = Checkbox(
        False,
        description=f"Remove {run_manager.move_old_logs_to}/ folder",
        indent=False,
    )
    confirm_cleanup_button = Button(
        description="Confirm", button_style="success", icon="check"
    )
    deny_cleanup_button = Button(
        description="Deny", button_style="danger", icon="close"
    )

    cleanup_box = VBox(
        [HBox([confirm_cleanup_button, deny_cleanup_button]), include_old_logs]
    )
    cleanup_button.on_click(switch_to(widgets["cleanup"], cleanup_box))
    deny_cleanup_button.on_click(switch_to(widgets["cleanup"], cleanup_button))
    confirm_cleanup_button.on_click(
        switch_to(
            widgets["cleanup"], cleanup_button, _callable=cleanup(include_old_logs)
        )
    )

    box.children = (status, *tuple(widgets.values()))
    display(box)
